import json
from functools import lru_cache
from pathlib import Path

data_dir = Path(__file__).parent.joinpath("data")


@lru_cache()
def load_defaults(version):
    version_file = data_dir.joinpath(f"{version}.json")
    if not version_file.exists():
        # Fall back to 950 defaults
        version_file = data_dir.joinpath("950.json")
    with version_file.open() as infile:
        return json.load(infile)



"""Top level API for parsing log files.

Usage example:
    import grblogtools.api as glt
    result = glt.parse("data/*.log")
    result.summary()
    result.progress(section="nodelog")

OR, use
    summary = glt.get_dataframe("data/*.log", timeline=False)
    summary, timeline = glt.get_dataframe("data/*.log", timeline=True)
"""

import glob
import itertools
from pathlib import Path
from typing import List, Union

import pandas as pd

from grblogtools.helpers import (
    add_categorical_descriptions,
    fill_default_parameters_nosuffix,
    strip_model_and_seed,
)
from grblogtools.parsers.single_log import SingleLogParser


class ParseResult:
    def __init__(self):
        self.parsers = []
        self._common = None

    def progress(self, section="nodelog") -> dict:
        """Return the search progress for the given section in the log.

        Args:
            section (str): Possible values are norel, rootlp, and nodelog. Defaults
                to nodelog.

        Returns:
            pd.DataFrame: A data frame representing the progress of the given section
                in the log.
        """
        progress = []
        for logfile, lognumber, parser in self.parsers:

            if section == "nodelog":
                log = parser.nodelog_parser.get_progress()
            elif section == "rootlp":
                log = parser.continuous_parser.get_progress()
            elif section == "norel":
                log = parser.norel_parser.get_progress()
            else:
                raise ValueError(f"Unknown section '{section}'")

            progress.append(
                pd.DataFrame(log).assign(LogFilePath=logfile, LogNumber=lognumber)
            )

        return pd.merge(
            left=pd.concat(progress),
            right=self.common_log_data(),
            how="left",
            on=["LogFilePath", "LogNumber"],
        )

    def common_log_data(self):
        """Extract common data to be joined to progress and summary dataframes.

        This could be cached in future and invalidated by .parse().
        """
        common = pd.DataFrame(
            [
                {
                    "LogFilePath": logfile,
                    "LogNumber": lognumber,
                    "ModelFilePath": parser.header_parser.get_summary().get(
                        "ModelFilePath"
                    ),
                    "Seed": parser.header_parser.get_parameters().get("Seed", 0),
                    "Version": parser.header_parser.get_summary().get("Version"),
                }
                for logfile, lognumber, parser in self.parsers
            ]
        )
        common = common.dropna(axis="columns", how="all")
        if "ModelFilePath" in common:
            common = common.assign(
                ModelFile=lambda df: df["ModelFilePath"].apply(
                    lambda p: None if p is None else Path(p).parts[-1].partition(".")[0]
                ),
                Model=lambda df: df["ModelFile"],
                Log=lambda df: df.apply(strip_model_and_seed, axis=1),
            )
        return common

    def summary(self, prettyparams=False):
        """Construct and return a summary dataframe for all parsed logs."""
        summary = pd.DataFrame(
            [
                dict(parser.get_summary(), LogFilePath=logfile, LogNumber=lognumber)
                for logfile, lognumber, parser in self.parsers
            ]
        )
        parameters = pd.DataFrame(
            [parser.header_parser.get_parameters() for _, _, parser in self.parsers]
        )
        # Fill defaults and add suffix to parameter columns.
        parameters = (
            fill_default_parameters_nosuffix(parameters.join(summary["Version"]))
            .drop(columns=["Version", "Seed"], errors="ignore")
            .rename(columns=lambda c: c + " (Parameter)")
        )
        # Convert parameters to categorical if required.
        if prettyparams:
            parameters = add_categorical_descriptions(parameters)
        summary = summary.join(parameters)
        summary = pd.merge(
            left=summary.drop(columns=["ModelFilePath", "Version"], errors="ignore"),
            right=self.common_log_data(),
            how="left",
            on=["LogFilePath", "LogNumber"],
        )
        return summary

    def parse(self, logfile: str) -> None:
        """Parse a single file. The log file may contain multiple run logs."""
        parser = SingleLogParser()
        subsequent = SingleLogParser()
        lognumber = 1
        with open(logfile) as infile:
            lines = iter(infile)
            for line in lines:
                if not parser.parse(line):
                    assert not subsequent.started
                    if subsequent.parse(line):
                        # The current parser did not match but an empty parser
                        # matched a header line.
                        self.parsers.append((logfile, lognumber, parser))
                        lognumber += 1
                        parser = subsequent
                        subsequent = SingleLogParser()

        self.parsers.append((logfile, lognumber, parser))


def parse(patterns: Union[str, List[str]]) -> ParseResult:
    """Main entry point function.

    Args:
        patterns (str): a single glob pattern, or list of patterns, matching
        log files.

    """
    result = ParseResult()
    if type(patterns) is str:
        patterns = [patterns]
    logfiles = itertools.chain(*(glob.glob(pattern) for pattern in patterns))
    for logfile in sorted(set(logfiles)):
        result.parse(logfile)
    return result


def get_dataframe(logfiles: List[str], timelines=False, prettyparams=False):
    """Compatibility function for the legacy API.

    If one log file contains more than one run, all runs are parsed, each reported
    as a separate row in the summary and timelines dataframes.

    Args:
        logfiles (str): A list of glob patterns of log files to be parsed.
        timelines (bool, optional): Return the norel, the relaxation, and the
            search tree progress if set to True. Defaults to False.
        prettyparams (bool, optional): Replace some parameter values with
            categorical labels.
    """
    result = parse(logfiles)
    summary = result.summary(prettyparams=prettyparams)
    if not timelines:
        return summary
    return summary, dict(
        norel=result.progress("norel"),
        rootlp=result.progress("rootlp"),
        nodelog=result.progress("nodelog"),
    )



import re

from grblogtools.parsers.util import typeconvert_groupdict


class SimplexParser:
    # The pattern indicating the initialization of the parser
    simplex_start_pattern = re.compile(
        r"Iteration(\s+)Objective(\s+)Primal Inf.(\s+)Dual Inf.(\s+)Time"
    )

    # The pattern indicating the simplex progress
    simplex_progress_pattern = re.compile(
        r"\s*(?P<Iteration>\d+)\s+(?P<Objective>[^\s]+)\s+(?P<PInf>[^\s]+)\s+(?P<DInf>[^\s]+)\s+(?P<Time>\d+)s"
    )

    # The pattern indicating the termination of the simplex method
    simplex_termination_pattern = re.compile(
        r"(Solved|Stopped) in (?P<IterCount>[^\s]+) iterations and (?P<Runtime>[^\s]+) seconds"
    )

    def __init__(self):
        """Initialize the Simplex parser."""
        self._summary = {}
        self._progress = []
        self._started = False

    def parse(self, line: str) -> bool:
        """Parse the given log line to populate summary and progress data.

        Args:
            line (str): A line in the log file.

        Returns:
            bool: Return True if the given line is matched by some pattern.
        """
        # Check this first since the termination line might appear
        # without any progress log in the concurrent case.
        match = SimplexParser.simplex_termination_pattern.match(line)
        if match:
            self._summary.update(typeconvert_groupdict(match))
            return True

        if not self._started:
            match = SimplexParser.simplex_start_pattern.match(line)
            if match:
                self._started = True
                return True
            return False

        progress_match = SimplexParser.simplex_progress_pattern.match(line)
        if progress_match:
            entry = {"Type": "simplex"}
            entry.update(typeconvert_groupdict(progress_match))
            self._progress.append(entry)
            return True

        return False

    def get_summary(self) -> dict:
        """Return the current parsed summary."""
        return self._summary

    def get_progress(self) -> list:
        """Return the detailed progress in simplex method if exists."""
        return self._progress



import argparse

import pandas as pd


def cli(glt_parse, argparse_kwargs):
    """Entry point function for command line interface

    Args:
        glt_parse: the parse function from grblogtools (in case overloads are needed)
        argparse_kwargs: extra info to pass to argparse

    """
    parser = argparse.ArgumentParser(**argparse_kwargs)
    parser.add_argument("outfile", help="Output file name (.xlsx)", metavar="XLSXFILE")
    parser.add_argument(
        "logfiles", help="Gurobi log files", nargs="+", metavar="LOGFILE"
    )
    parser.add_argument(
        "-t",
        "--timelines",
        action="store_true",
        help="also store timelines (root LP, node log, and NoRel log) in separate sheets",
    )
    args = parser.parse_args()

    result = glt_parse(args.logfiles)
    summary = result.summary()

    with pd.ExcelWriter(args.outfile) as writer:
        summary.to_excel(writer, sheet_name="Summary")
        for tl in ["norel", "rootlp", "nodelog"]:
            result.progress(tl).to_excel(writer, sheet_name=tl)

    print(f"extracted {len(summary)} log(s) to {args.outfile}")



import re

from grblogtools.parsers.util import convert_data_types, typeconvert_groupdict


class HeaderParser:
    header_start_patterns = [
        re.compile(
            r"Gurobi (?P<Version>\d{1,2}\.[^\s]+) \((?P<Platform>[^\)]+)\) logging started (?P<Time>.*)$"
        ),
        re.compile(r"Logging started (?P<Time>.*)$"),
        re.compile(
            r"Gurobi Compute Server Worker version (?P<Version>\d{1,2}\.[^\s]+) build (.*) \((?P<Platform>[^\)]+)\)$"
        ),
        re.compile(r"Compute Server job ID: (?P<JobID>.*)$"),
        re.compile(r"Gurobi Optimizer version (?P<Version>\d{1,2}\.[^\s]+)"),
        re.compile(
            r"Gurobi Compute Server Worker version (?P<Version>\d{1,2}\.[^\s]+) build (.*) \((?P<Platform>[^\)]+)\)$"
        ),
        re.compile(r"Compute Server job ID: (?P<JobID>.*)$"),
        re.compile(r"Gurobi Optimizer version (?P<Version>\d{1,2}\.[^\s]+)"),
    ]

    header_other_patterns = [
        re.compile(r"Read (MPS|LP) format model from file (?P<ModelFilePath>.*)$"),
        re.compile(r"Reading time = (?P<ReadingTime>[\d\.]+) seconds"),
        re.compile(
            r"Thread count: (?P<PhysicalCores>\d+) physical cores, (?P<LogicalProcessors>\d+) logical processors, using up to (?P<Threads>\d+) threads"
        ),
        re.compile(
            r"(?P<ModelName>.*): (?P<Rows>\d+) rows, (?P<Columns>\d+) columns, (?P<Nonzeros>\d+) nonzeros"
        ),
    ]

    # Special case for parameter changes
    parameter_change_pattern = re.compile(
        r"Set parameter (?P<ParamName>[^\s]+) to value (?P<ParamValue>.*)$"
    )

    def __init__(self):
        """Initialize the Header parser.

        Parameters are stored separately from the summary data as they are
        handled differently in the final output.
        """
        self._summary = {}
        self._parameters = {}
        self._started = False

    def parse(self, line: str) -> bool:
        """Parse the given log line to populate summary data.

        Args:
            line (str): A line in the log file.

        Returns:
            bool: Return True if the given line is matched by some pattern.
        """
        match = HeaderParser.parameter_change_pattern.match(line)
        if match:
            self._parameters[match.group("ParamName")] = convert_data_types(
                match.group("ParamValue")
            )
            return True

        for pattern in HeaderParser.header_start_patterns:
            match = pattern.match(line)
            if match:
                self._started = True
                self._summary.update(typeconvert_groupdict(match))
                return True

        if self._started:
            for pattern in HeaderParser.header_other_patterns:
                match = pattern.match(line)
                if match:
                    self._summary.update(typeconvert_groupdict(match))
                    return True

        return False

    def get_summary(self) -> dict:
        """Return the current parsed summary."""
        return self._summary

    def get_parameters(self) -> dict:
        """Return all changed parameters detected in the header."""
        return self._parameters



import re

from grblogtools.parsers.util import typeconvert_groupdict


class TerminationParser:
    # Termination patterns
    patterns = [
        re.compile(r"ERROR (?P<ErrorCode>[^:]+): (?P<ErrorMessage>.*)$"),
        re.compile(r"\[(?P<ErrorMessage>process terminated with exit code [^\\]]+)\]$"),
        re.compile(r"(?P<TIME_LIMIT>Time limit reached)"),
        re.compile(r"(?P<OPTIMAL>Optimal solution found)(?: \(tolerance .*\))"),
        re.compile(r"(?P<ITERATION_LIMIT>Iteration limit reached)"),
        re.compile(r"(?P<INF_OR_UNBD>Infeasible or unbounded model)"),
        re.compile(r"(?P<INF_OR_UNBD>Model is infeasible or unbounded)"),
        re.compile(r"(?P<UNBOUNDED>Unbounded model)"),
        re.compile(r"(?P<UNBOUNDED>Model is unbounded)"),
        re.compile(r"(?P<INFEASIBLE>Infeasible model)"),
        re.compile(r"(?P<INFEASIBLE>Model is infeasible)"),
        re.compile(r"(?P<SOLUTION_LIMIT>Solution limit reached)"),
        re.compile(r"(?P<NODE_LIMIT>Node limit reached)"),
        re.compile(r"(?P<NUMERIC>Numeric error)"),
        re.compile(r"(?P<NUMERIC>Numerical trouble encountered)"),
        re.compile(r"(?P<CUTOFF>Model objective exceeds cutoff)"),
        re.compile(r"(?P<CUTOFF>Objective cutoff exceeded)"),
        re.compile(r"(?P<USER_OBJ_LIMIT>Optimization achieved user objective limit)"),
        re.compile(
            r"(?P<INTERRUPTED>(Interrupt request received|Solve interrupted))(?: \\(error code (?P<ErrorCode>[^\\)]+)\\))?"
        ),
        re.compile(r"Solution count (?P<SolCount>\d+)"),
        re.compile(
            r"Thread count was (?P<Threads>\d+) \(of (?P<Cores>\d+) available processors\)"
        ),
    ]

    status = [
        "OPTIMAL",
        "TIME_LIMIT",
        "ITERATION_LIMIT",
        "INF_OR_UNBD",
        "UNBOUNDED",
        "INFEASIBLE",
        "SOLUTION_LIMIT",
        "NUMERIC",
        "CUTOFF",
        "USER_OBJ_LIMIT",
        "INTERRUPTED",
    ]

    def __init__(self):
        self._summary = {}

    def parse(self, line: str) -> bool:
        """Return True if the line is matched by some pattern.

        Args:
            line (str): A line in the log file.

        Returns:
            bool: Return True if the given line is matched by some pattern.
        """
        for pattern in TerminationParser.patterns:
            match = pattern.match(line)
            if match:
                for key, value in typeconvert_groupdict(match).items():
                    if key in TerminationParser.status:
                        self._summary.update({"Status": key})
                    else:
                        self._summary.update({key: value})
                return True
        return False

    def get_summary(self) -> dict:
        """Return the current parsed summary."""
        return self._summary
